// -*- C++ -*-
#include "Rivet/Projections/SubtractedJewelEvent.hh"

namespace Rivet {

  void SubtractedJewelEvent::project(const Event& e) {
    clear();
    translateParticles(e);
    vector<MyDist> dists = buildPairs();
    doSubtraction(dists);
    collectResult();
    return;
  }

  void SubtractedJewelEvent::clear() {
    _subtrevent = Particles();
    _myparticles.clear();
    _mythermals.clear();
  }

void SubtractedJewelEvent::translateParticles(const Event& e) {
	//translate Particle's and FourMomentum's into myPart's 
	for (const HepMC::GenParticle* p : Rivet::HepMCUtils::particles(e.genEvent())) {
	    if(p->status() == 1 || p->status() == 4 ) {
			FourMomentum partmom = p->momentum();
			if (fuzzyEquals(partmom.E(),1e-6)) continue;
			if (partmom.E() - fabs(partmom.pz()) > 0.) {
				MyPart part;
				part.id = p->pdg_id();
				part.pt = partmom.pT();
				part.mdelta = sqrt(partmom.mass2() + sqr(partmom.pT())) - partmom.pT();
				part.phi = partmom.p3().phi();
				part.y = partmom.rapidity();
				_myparticles.push_back(part);
			}
			else {
				_subtrevent.push_back(p);
			}
	    }
	    
	    if (p->status() == 3) {
		    FourMomentum mom = p->momentum();
			My4Mom therm;
			if (mom.E() - fabs(mom.pz()) > 0.) {
				therm.pt = mom.pT();
				therm.mdelta = sqrt(mom.mass2() + sqr(mom.pT())) - mom.pT();
				therm.phi = mom.p3().phi();
				therm.y = mom.rapidity();
				_mythermals.push_back(therm);
			}
			else {
				PdgId thermid;
				if (rand()/RAND_MAX < 1./3.) thermid = PID::PIPLUS;
				else if (rand()/RAND_MAX < 2./3.) thermid = PID::PIMINUS;
				else thermid = PID::PI0;
				Particle part(thermid,mom);
				_subtrevent.push_back(part);
			}
	    }
	}
}
    
vector<MyDist> SubtractedJewelEvent::buildPairs() {
    //create list with all particle-ghosts distances and sort it
	  vector<MyDist> dists;
	  for (size_t i=0; i < _myparticles.size(); ++i) {
		  for (size_t j=0; j < _mythermals.size(); ++j) {
			  //cout<<"i, j = "<<i<<" "<<j<<endl;
			  MyDist dist;
			  double Deltaphi(abs(_myparticles[i].phi-_mythermals[j].phi));
			  if (Deltaphi > M_PI) Deltaphi = 2.*M_PI - Deltaphi;
			  dist.dR = sqrt(sqr(Deltaphi) + sqr(_myparticles[i].y-_mythermals[j].y));
			  dist.ipart = i;
			  dist.itherm = j;
			  //cout<<"distance: "<<dist.dR<<endl;
			  dists.push_back(dist);
		  }
	  }
	  //dists.sort(MyDistComp);
	  std::sort(dists.begin(),dists.end(),MyDistComp);
	  return dists;
}
    
void SubtractedJewelEvent::doSubtraction(vector<MyDist> & dists) {
		  //go through all particle-ghost pairs and re-distribute momentum and mass
	  for (vector<MyDist>::iterator liter=dists.begin(); liter != dists.end(); ++liter) {
		  //cout<<"dealing with dist "<<liter->dR<<endl;
		  if (liter->dR > _dRmax) break;
		  size_t pnum = liter->ipart;
		  size_t tnum = liter->itherm;
		  double ptp = _myparticles[pnum].pt;
		  double ptt = _mythermals[tnum].pt;
		  //cout<<"pts: "<<ptp<<" vs "<<ptg<<endl;
		  if (ptp > ptt) {
			  _myparticles[pnum].pt -= ptt;
			  _mythermals[tnum].pt = 0.;
		  }
		  else {
			  _mythermals[tnum].pt -= ptp;
			  _myparticles[pnum].pt = 0.;
		  }
		  double mdp = _myparticles[pnum].mdelta;
		  double mdt = _mythermals[tnum].mdelta;
		  //cout<<"masses: "<<mdp<<" vs "<<mdg<<endl;
		  if (mdp > mdt) {
			  _myparticles[pnum].mdelta -= mdt;
			  _mythermals[tnum].mdelta = 0.;
		  }
		  else {
			  _mythermals[tnum].mdelta -= mdp;
			  _myparticles[pnum].mdelta = 0.;
		  }
		  
	  }
	  return;
}
    
void SubtractedJewelEvent::collectResult() {
		  //collect resulting 4-momenta to get subtracted event
	  for (size_t i=0; i < _myparticles.size(); ++i) {
		  if (_myparticles[i].pt > 0.) {
			  FourMomentum mom((_myparticles[i].pt+_myparticles[i].mdelta)*cosh(_myparticles[i].y), 
				_myparticles[i].pt*cos(_myparticles[i].phi),
				_myparticles[i].pt*sin(_myparticles[i].phi),
				(_myparticles[i].pt+_myparticles[i].mdelta)*sinh(_myparticles[i].y));
			  Particle outpart(_myparticles[i].id,mom);
			  _subtrevent.push_back(outpart);
		  }
	  }
	  for (size_t i=0; i < _mythermals.size(); ++i) {
		  if (_mythermals[i].pt > 0.) {
			  FourMomentum mom((_mythermals[i].pt+_mythermals[i].mdelta)*cosh(_mythermals[i].y), 
				_mythermals[i].pt*cos(_mythermals[i].phi),
				_mythermals[i].pt*sin(_mythermals[i].phi),
				(_mythermals[i].pt+_mythermals[i].mdelta)*sinh(_mythermals[i].y));
			  PdgId thermid;
			  if (rand()/RAND_MAX < 1./3.) thermid = PID::PIPLUS;
			  else if (rand()/RAND_MAX < 2./3.) thermid = PID::PIMINUS;
			  else thermid = PID::PI0;
			  Particle outpart(thermid,mom);
			  _subtrevent.push_back(outpart);
		}
	  }
        return;
}

}
    
